"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const cryptography = __importStar(require("@liskhq/lisk-cryptography"));
const browserify_bignum_1 = __importDefault(require("browserify-bignum"));
const constants_1 = require("../../constants");
const MAX_PUBLIC_KEY_LENGTH = 32;
exports.validatePublicKey = (publicKey) => {
    const publicKeyBuffer = cryptography.hexToBuffer(publicKey);
    if (publicKeyBuffer.length !== MAX_PUBLIC_KEY_LENGTH) {
        throw new Error(`Public key ${publicKey} length differs from the expected 32 bytes for a public key.`);
    }
    return true;
};
exports.checkPublicKeysForDuplicates = (publicKeys) => publicKeys.every((element, index) => {
    if (publicKeys.slice(index + 1).includes(element)) {
        throw new Error(`Duplicated public key: ${publicKeys[index]}.`);
    }
    return true;
});
exports.validatePublicKeys = (publicKeys) => publicKeys.every(exports.validatePublicKey) &&
    exports.checkPublicKeysForDuplicates(publicKeys);
exports.validateKeysgroup = (keysgroup) => {
    if (keysgroup.length < constants_1.MULTISIGNATURE_MIN_KEYSGROUP ||
        keysgroup.length > constants_1.MULTISIGNATURE_MAX_KEYSGROUP) {
        throw new Error(`Expected between ${constants_1.MULTISIGNATURE_MIN_KEYSGROUP} and ${constants_1.MULTISIGNATURE_MAX_KEYSGROUP} public keys in the keysgroup.`);
    }
    return exports.validatePublicKeys(keysgroup);
};
const MIN_ADDRESS_LENGTH = 2;
const MAX_ADDRESS_LENGTH = 22;
const BASE_TEN = 10;
exports.validateAddress = (address) => {
    if (address.length < MIN_ADDRESS_LENGTH ||
        address.length > MAX_ADDRESS_LENGTH) {
        throw new Error('Address length does not match requirements. Expected between 2 and 22 characters.');
    }
    if (address[address.length - 1] !== 'L') {
        throw new Error('Address format does not match requirements. Expected "L" at the end.');
    }
    if (address.includes('.')) {
        throw new Error('Address format does not match requirements. Address includes invalid character: `.`.');
    }
    const addressString = address.slice(0, -1);
    const addressNumber = new browserify_bignum_1.default(addressString);
    if (addressNumber.cmp(new browserify_bignum_1.default(constants_1.MAX_ADDRESS_NUMBER)) > 0) {
        throw new Error('Address format does not match requirements. Address out of maximum range.');
    }
    if (addressString !== addressNumber.toString(BASE_TEN)) {
        throw new Error("Address string format does not match it's number representation.");
    }
    return true;
};
exports.isGreaterThanZero = (amount) => amount.cmp(0) > 0;
exports.isGreaterThanMaxTransactionAmount = (amount) => amount.cmp(constants_1.MAX_TRANSACTION_AMOUNT) > 0;
exports.isGreaterThanMaxTransactionId = (id) => id.cmp(constants_1.MAX_TRANSACTION_ID) > 0;
exports.isNumberString = (str) => {
    if (typeof str !== 'string') {
        return false;
    }
    return /^[0-9]+$/g.test(str);
};
exports.validateAmount = (data) => exports.isNumberString(data) && !exports.isGreaterThanZero(new browserify_bignum_1.default(data));
exports.validateTransferAmount = (data) => exports.isNumberString(data) &&
    exports.isGreaterThanZero(new browserify_bignum_1.default(data)) &&
    !exports.isGreaterThanMaxTransactionAmount(new browserify_bignum_1.default(data));
exports.validateFee = (data) => exports.isNumberString(data) &&
    exports.isGreaterThanZero(new browserify_bignum_1.default(data)) &&
    !exports.isGreaterThanMaxTransactionAmount(new browserify_bignum_1.default(data));
exports.isValidInteger = (num) => typeof num === 'number' ? Math.floor(num) === num : false;
//# sourceMappingURL=validation.js.map